{Start of file MultiID.Pas *************************************************}

unit MultiID; {Multiple-precision integer decimal algorithms}

{ Version = 'Version 2.50, last revised: 1994-11-20, 1100 hours';
  Author  = 'Copyright (c) 1981-1994 by author: Harry J. Smith,';
  Address = '19628 Via Monte Dr., Saratoga, CA 95070.  All rights reserved.';
}
{***************************************************************************}

{$I-} {Do our own i/o error checks}
{$N+} {Uses numeric coprocessor}
{$R-} {No index Range checking}
{$V-} {Parameter string length need not match}

{Developed in TURBO Pascal 5.5, maintained in TURBO Pascal 6.0}

{ Revisions made - Listed after end. statement }

{In Turbo Pascal 4.0 etc. the max odd integer to store in a Single is}
{2**24-1 = 167,77215; for a Double is 2**53-1 = 9,00719,92547,40991}

interface

uses Dos, Crt, Printer; {Turbo Pascal interface}

const
  MuDMax = 7;     {Max number of decimal digits in a super digit}
  MuNMax = 16379; {Max number of super digits in a multi-precision number}
                  {Each super digit is from 0 to 9999999, i.e. Base = 10**7}
type
  Mu1Digit = Single; {Can hold + or - integer <= Base}
  Mu2Digit = Double; {Can hold + or - integer <= Base squared + Base}
  VType    = Array[0..MuNMax] of Mu1Digit; {Value, LSD at index 0}
  VPtr     = ^VType;

  MultiIDO = object {Multiple-precision integer decimal algorithms}
    constructor Init( MuMbM : Integer);
    destructor Done;
  end;

  MultiUI = object {A multiple-precision unsigned integer}
    N : Integer; {Current length in super digits, 1..MuNMax}
    M : Integer; {Max number of super digits ever in this number}
    V : VPtr;
    constructor Init( NMaxI : Integer);
    destructor Done; virtual;
    procedure Clear; virtual; {Set value = 0}
    procedure SetTo( var X : MultiUI); {Self:= X}
    procedure Norm; {Normalize}
    procedure RAdd( var X : MultiUI); {Self:= Self + X}
    procedure RSub( var X : MultiUI); {Self:= Self - X}
    procedure Add( var X, Y : MultiUI); {Self:= X + Y}
    procedure Sub( var X, Y : MultiUI); {Self:= X - Y}
    function  AbComp( var X : MultiUI): Integer; {AbComp = sign(|Self| - |X|)}
    function  ZTest : Boolean; {True if Self is Zero}
    function  EQ1( D1 : Mu1Digit) : Boolean; {True if |Self| = D1}
    function  GT1( D1 : Mu1Digit) : Boolean; {True if |Self| > D1}
  end;

  MultiSI = object( MultiUI) {A multiple-precision signed integer}
    S : Boolean; {Sign, True = negative}
    constructor Init( NMaxI : Integer);
    destructor Done; virtual;
    procedure Clear; virtual; {Set value = 0}
    procedure ChS; {Change sign, Self = 0 - Self}
    procedure SetTo( var X : MultiSI); {Self:= X}
    procedure Norm; {Normalize}
    procedure RAdd( var X : MultiSI); {Self:= Self + X}
    procedure RSub( var X : MultiSI); {Self:= Self - X}
    procedure Add( var X, Y : MultiSI); {Self:= X + Y}
    procedure Sub( var X, Y : MultiSI); {Self:= X - Y}
    procedure Value( St : String; var I : Integer);
      {Convert String to Self, returns I = # of character used}
    procedure SetTo1( D1 : Mu1Digit); {Self = D1 Mod Base}
    procedure SetToEx( Ex : Extended); {Self = Ex Mod Base**4}
    procedure SetToD( Db : Double); {Self = Db Mod Base**4}
    procedure Get1( var D1 : Mu1Digit); {D1 = Self Mod Base}
    procedure GetEx( var Ex : Extended); {Ex = Self Mod Base**4}
    procedure GetD( var Db : Double); {Db = Self Mod Base**4}
    procedure RAdd1( D1 : Mu1Digit); {Self = Self + D1}
    procedure Add1( var X : MultiSI; D1 : Mu1Digit); {Self = X + D1}
    function  Comp( var X : MultiSI) : Integer; {Comp = sign( Self - X)}
    procedure Writ( var Out : TEXT); {Output Self as a line of text}
    procedure WritLn( var Out : TEXT); {Output Self and a new line}
    procedure ShortWr( var Out : TEXT; Short : LongInt);
      {Output Self in short form ... if more than Short digits}
    procedure ShortWrLn( var Out : TEXT; Short : LongInt);
      {Output Self short and a new line}
    procedure RMul1( D1 : Mu1Digit); {Self = Self * D1}
    procedure Mul1( var X : MultiSI; D1 : Mu1Digit); {Self = X * D1}
    procedure Mul( var X, Y : MultiSI); {Self = X * Y}
    procedure RMul( var X : MultiSI); {Self = Self * X}
    procedure Sq( var X : MultiSI); {Self = X * X}
    procedure RSq; {Self = Self * Self}
    procedure PowMB( var B, P : MultiSI); {Self = (B ** P) Mod MuMB}
    procedure RPowMB( var P : MultiSI); {Self = (Self ** P) Mod MuMB}
    procedure ModMB; {Self = Self Mod MuMB}
    procedure RDiv1( D1 : Mu1Digit); {Self = Self / D1}
    procedure Div1( var U : MultiSI; D1 : Mu1Digit); {Self = U / D1}
    procedure Mod1( D1: Mu1Digit; var R1: Mu1Digit); {R1 = Rem( Self / D1)}
    procedure RDiv1QR( D1: Mu1Digit; var R1: Mu1Digit);
      {Self = Self / D1, R1 = Rem}
    procedure Div1QR( var U: MultiSI; D1: Mu1Digit; var R1: Mu1Digit);
      {Self = U / D1, R1 = Rem}
    procedure Divi( var U, D : MultiSI); {Self = U / D}
    procedure RDiv( var D : MultiSI); {Self = Self / D}
    procedure Modu( var U, D : MultiSI); {Self = Rem(U / D)}
    procedure RMod( var D : MultiSI); {Self = Rem( Self / D)}
    procedure DivQR( var U, D, R : MultiSI); {Self = U / D, R = Rem}
    procedure RDivQR( var D, R : MultiSI); {Self = Self / D, R = Rem}
    procedure GCD( var A, B : MultiSI);{Self = Greatest common divisor A, B}
    procedure RGCD( var A : MultiSI); {Self=Greatest common divisor A, Self}
    procedure FacMB( var X : MultiSI); {Self = (X !) Mod MuMB = 1*2*3*...*X}
    procedure RFacMB; {Self = (Self !) Mod MuMB = 1*2*3*...*X}
    procedure SqRtRem( var X, R : MultiSI); {Self = SqRt(X), R = Rem}
    procedure RSqRtRem( var R : MultiSI); {Self = SqRt( Self), R = Rem}
    procedure SqRoot( var X : MultiSI); {Self = SqRt(X)}
    procedure RSqRt; {Self = SqRt( Self)}
  end;

var
  MultiIDV : MultiIDO;    {Multiple-precision integer decimal algorithms}

  {The following are set when the unit MultiID is loaded}
  Base     : Mu2Digit;    {Radix of multi-precision numbers, 10**MuDMax}
  MaxDigit : Mu1Digit;    {Base - 1}
  BaseSq   : Mu2Digit;    {Base squared}
  BSMB     : Mu2Digit;    {(BaseSq - Base) for MuDiv and MuSqRt}
  OutP     : Array[0..1] of ^Text; {Pointers to Output and Lst}
                          {The following can be tested by user}
  MuErr    : Boolean;     {True if multi-precision error occurred}
  MuAbort  : Boolean;     {True if Mu procedure interrupted by operator}
  SoftAbort: Boolean;     {True if SoftAbort Flag set by operator}
  KeyHit   : Boolean;     {True if KeyPressed has been cleared}
  KeyCh    : Char;        {Key input when KeyPressed was cleared}
                          {The following can be changed by user}
  Echo     : Integer;     {1 if in Echo output to printer mode, 0 otherwise}
  MuDpG    : Integer;     {Digits per group in number output display >= 2}
  MuLenD   : Integer;     {Min length in digits to display length in digits}
  MuErrRep : Boolean;     {True if multi-precision error reports are desired}
  MMax     : Integer;     {Max number of super digits <= MuNMax}
  TracN    : LongInt;     {Number if iterations of an innermost loop}
  Trace    : LongInt;     {Number of "Digits" left to compute}
  DiagOn   : Boolean;     {Diagnostics turned on}
  TV0      : Double;      {Dos time value at time zero for diag}
  TV1      : Double;      {Dos time value for delta time, TV2 - TV1}
  TV2      : Double;      {Dos time value for total time, TV2 - TV0}
  DosClockP: Double;      {Previous value of the Dos clock for day cycle test}
  DosDays  : Integer;     {Number of days to add to Dos clock}
  MuMaxW   : Integer;     {Max number of characters to write on a line}
  MuTot    : Integer;     {Total number of characters on the line so far}

  {The following are set when the object MultiIDV is initialized}
  MuMB     : ^MultiSI;    {Multi-precision modulo base, 0 => normal aritm.}

procedure ShiftUp( var St : String); {Shift string to upper case}
procedure MuWriteErr( St : String); {Write error statement and set error flag}
procedure ReadLInt( Mess : String; Min, Max, Nom : LongInt;
                 var II : LongInt);   {Read in a long integer from keyboard}
procedure ReadInt( Mess  : String; Min, Max, Nom : Integer;
                 var II : Integer);   {Read in an integer from keyboard}
procedure MuSetMMax( NRegs : Integer); {Set MMax using MemAvail}
procedure MuInterrupt; {Test for ESC key pressed to abort operation}
function  DosClock : Double; {Get the value of the Dos clock in total seconds}
                  {Only good to 0.01 seconds, must call earlier once each day}
procedure Diag( Mess : String); {Output a diagnostic message with delta times}
procedure WriteMax( var Out : TEXT; St : String);
{Output String and line feed every MuMaxW, MuTot = characters on line so far}
procedure WriteMaxLn( var Out : TEXT; St : String);

implementation

{--------------------------------------}
{       Common hidden routines         }
{--------------------------------------}

{--------------------------------------}
procedure IpAdd( var X, Y : MultiUI; var K : Mu1Digit); {|Y| = |Y| + |X|}
{Multi-precision integer add absolute values in place;  Y is changed}
{Y.N Must be >= X.N, X.S and Y.S not used or changed;  X not changed}
{Overflow super digit returned not zero if and only if Y overflows}
{MuErr flag not set.  Address of X and Y can be the same, then X is changed}
var
  I : Integer;
  T : Mu2Digit;
begin
  I:= 0;  K:= 0;
  while I < X.N do begin
    T:= X.V^[I] + Y.V^[I] + K; {Add next super digit with carry}
    if T >= Base then begin
      K:= 1;  Y.V^[I]:= T - Base;
    end
    else begin
      K:= 0;  Y.V^[I]:= T;
    end;
    Inc(I);
  end;
  if K = 0 then  Exit;
  while (I < Y.N) do begin    {X = 0 from here on}
    if Y.V^[I] = MaxDigit then begin {K = 1, carry to next super digit of Y}
      Y.V^[I]:= 0;
    end
    else begin
      Y.V^[I]:= Y.V^[I] + 1;  K:= 0;  Exit; {Return K = 0}
    end;
    Inc(I);
  end;
  Y.N:= Y.N + 1;
  if Y.N > Y.M then begin {If overflow return K = 1}
    Y.N:= Y.M;
    Y.Norm;
  end
  else begin
    Y.V^[ Y.N-1]:= 1;  K:= 0; {Return K = 0}
  end;
end; {IpAdd}

{--------------------------------------}
procedure IpSub( var X, Y : MultiUI); {|Y| = |Y| - |X|}
{Multi-precision integer subtract absolute values in place;  Y is changed}
{|Y| Must be >= |X|, X.S and Y.S not used or changed;  X not changed}
{MuErr flag not set.  Address of X and Y can be the same, then X is changed}
var
  I : Integer;
  K : Mu1Digit;
  T : Mu2Digit;
begin
  I:= 0;  K:= 0;
  while (I < X.N) do begin
    T:= Y.V^[I] - X.V^[I] + K; {Subtract next super digit with borrow}
    if T < 0 then begin
      K:= -1;  Y.V^[I]:= T + Base;
    end
    else begin
      K:= 0;  Y.V^[I]:= T;
    end;
    Inc(I);
  end;
  if (K = 0) or (X.N = Y.N) then begin  Y.Norm;  Exit;  end;
  while (I < Y.N) do begin {X = 0 from here on}
    if Y.V^[I] = 0 then begin {K = -1, borrow from next super digit of Y}
      Y.V^[I]:= MaxDigit;
    end
    else begin
      Y.V^[I]:= Y.V^[I] - 1;  Y.Norm;  Exit;
    end;
    Inc(I);
  end;
end; {IpSub}

{--------------------------------------}
procedure IpMul1(D : Mu1Digit; var X : MultiUI; var K : Mu1Digit);{X = |D| * X}
{Multi-precision one super digit multiply in place;  X is changed}
{K = Overflow super digit returned > zero if and only if X overflows}
{Signs not used}
var
  I : Integer;
  T : Mu2Digit;
begin
  D:= Abs(D);
  K:= 0;
  for I:= 0 to X.N - 1 do begin
    T:= D * X.V^[I] + K; {Multiply next super digit and add carry}
    K:= Int(T / Base);
    X.V^[I]:= T - K * Base;
  end;
  if K = 0 then  Exit;
  X.N:= X.N + 1;
  if X.N > X.M then begin {If overflow return K > 0}
    X.N:= X.M;
  end
  else begin
    X.V^[ X.N-1]:= K;  K:= 0;
  end;
end; {IpMul1}

{--------------------------------------}
procedure IpDiv1(D : Mu1Digit; var U : MultiSI; var R : Mu1Digit); {U = U / D}
{Multi-precision one super digit divide in place, U is changed} {R = Remainder}
{Allows literal D, MuErr set True if D = 0}
var
  T  : Mu2Digit;
  I  : Integer;
  AD : Mu1Digit;
begin
  R:= 0;
  if D = 0 then begin
    MuWriteErr('Cannot divide by zero, continuing...');
    Exit;
  end;
  MuInterrupt;
  if MuAbort then begin
    U.Clear;  Exit;
  end;
  AD:= Abs(D);
  I:= U.N - 1;
  while I >= 0 do begin
    T:= R * Base + U.V^[I]; {Divide next super digit and save remainder}
    U.V^[I]:= Int(T / AD);
    R:= T - U.V^[I] * AD;
    Dec(I);
  end;
  if U.S and (R <> 0) then R:= -R;
  U.S:= U.S xor (D < 0);
  U.Norm;
end; {IpDiv1}

{--------------------------------------}
procedure IpDiv( var D, U, Q, R : MultiSI); {Q = U / D, R = Rem}
{U and D are not changed, U is moved to R and then R is divided}
{in place to give Q and a remainder in R, it is OK for U and R}
{to have the same location in memory, but then U will be changed}
{Sets MuErr True if D = 0}
{See Knuth Vol. 2 of "The Art..." Algorithm D, pg. 237.}
var
  V1, V2, D1, K, QH : Mu1Digit;
  T                : Mu2Digit;
  I, J, IJ         : Integer;
begin
  if D.N = 1 then begin {Use IpDiv1 for a one super digit divide}
    Q.SetTo(U);  D1:= D.V^[0];
    if D.S and (D1 <> 0) then D1:= -D1;
    IpDiv1( D1, Q, K);
    R.S:= (K < 0);
    R.V^[0]:= Abs(K);
    R.N:= 1;
  end
  else begin
    R.SetTo(U); {SetTo is No-Op if same address}
    if R.N < D.N then begin
      Q.Clear;  Exit;
    end;
    Q.N:= R.N - D.N + 1;
    J:= R.N;
    R.V^[ R.N]:= 0;
    D.V^[ D.N]:= 0;
    V1:= D.V^[ D.N-1];
    V2:= D.V^[ D.N-2];
    D1:= Int( Base / (V1 + 1));      {D1. Normalize}
    if D1 > 1 then begin
      IpMul1( D1, D, K);  V1:= D.V^[ D.N-1];  V2:= D.V^[ D.N-2];
      IpMul1( D1, R, K);
      if K <> 0 then R.V^[J]:= K;
    end;                            {D2. Init J done above}
    KeyHit:= False;
    while J >= D.N do begin         {D3. Calculate Q hat}
      T:= R.V^[J] * Base + R.V^[ J-1];
      if V1 = R.V^[J]
        then  QH:= MaxDigit
        else  QH:= Int(T / V1);
      while (V2 * QH > (T - QH * V1) * Base + R.V^[ J-2]) do QH:= QH - 1;
{QH:=QH+1; .Diag for add back test}
      K:= Base;
      for I:= 0 to D.N do begin     {D4. Multiply and subtract}
        IJ:= I + J - D.N;
        T:= R.V^[ IJ] - QH * D.V^[I] + K + BSMB; {+ BaseSq - Base}
        K:= Int(T / Base);
        R.V^[ IJ]:= T - K * Base;
      end;
      Q.V^[J - D.N]:= QH;           {D5. Test remainder}
      if K = MaxDigit then begin
{WriteLn('MuDiv: Doing an add back'); .Diag for add back test}
        Q.V^[J - D.N]:= QH - 1;     {D6. Add back}
        K:= 0;  IJ:= J - D.N;
        for I:= 0 to D.N do begin
          T:= D.V^[I] + R.V^[ IJ] + K;
          K:= Int(T / Base);
          R.V^[ IJ]:= T - K * Base;
          Inc( IJ);
        end;
      end;
      Dec(J);                       {D7. Loop on J}
      MuInterrupt;
      if MuAbort then begin
        Q.Clear;  R.Clear;  Exit;
      end;
      if KeyHit then begin
        for I:= 0 to Echo do
          WriteLn( OutP[I]^, 'Integer Divide: ',
            Trunc( (100.0 * (R.N - J)) / (R.N - D.N + 1)), ' Percent done');
        KeyHit:= False;
      end;
    end;
    Q.S:= U.S xor D.S;
    R.S:= U.S;
    Q.Norm;
    R.Norm;
    if D1 > 1 then begin            {D8. Unnormalize}
      IpDiv1( D1, R, K);
      IpDiv1( D1, D, K);
    end;
  end;
end; {IpDiv}

{---End of common hidden routines------}

{--------------------------------------}
{   MultiIDO's method implementation   }
{--------------------------------------}

{--------------------------------------}
constructor MultiIDO.Init( MuMbM: Integer);{Initialize Multi-precision package}
{Input: MuMbM, M for MuMb}
{Initialize: MuMB = 0}
begin
  if MuMbM > MuNMax then  MuMbM:= MuNMax;
  New( MuMB, Init( MuMbM));
end; {MultiIDO.Init}

{--------------------------------------}
destructor MultiIDO.Done; {Destruct Multi-precision package}
begin
  Dispose( MuMB, Done);
end; {MultiIDO.Done}

{------End of MultiIDO's methods-------}

{--------------------------------------}
{   MultiUI's method implementation    }
{--------------------------------------}
constructor MultiUI.Init( NMaxI : Integer);
var
  Size : Word;
  I    : Integer;
begin
  M:= NMaxI;
  if M > MuNMax then  M:= MuNMax;
  Size:= SizeOf( Mu1Digit);
  Size:= (M+1) * Size;
  if Size > MaxAvail then begin
    if MemAvail = MaxAvail then
      for I:= 0 to Echo do  Write( OutP[I]^, 'MemAvail = MaxAvail = ')
    else
      for I:= 0 to Echo do
        Write( OutP[I]^, 'MemAvail = ', MemAvail, ', MaxAvail = ');
    for I:= 0 to Echo do
      WriteLn( OutP[I]^, MaxAvail, ', need ', Size);
    M:= 1;  Size:= SizeOf( Mu1Digit);  Size:= (M+1) * Size;  MuAbort:= True;
  end;
  GetMem(V, Size);
  Clear;
end; {MultiUI.Init}

{--------------------------------------}
destructor MultiUI.Done;
var
  Size : Word;
begin
  Size:= SizeOf( Mu1Digit);
  Size:= (M+1) * Size;
  FreeMem(V, Size);
  M:= -1;
end; {MultiUI.Done}

{--------------------------------------}
procedure MultiUI.Clear; {Set value = 0}
begin
  V^[0]:= 0;  N:= 1;
end; {MultiUI.Clear}

{--------------------------------------}
procedure MultiUI.SetTo( var X : MultiUI); {Self = X, copy used "digits" only}
var
  I : Integer;
begin
  if Self.V = X.V then  Exit;
  N:= X.N;
  if N > M then begin
    N:= M;
    MuWriteErr('SetTo overflow, continuing...');
  end;
  for I:= 0 to N - 1 do  V^[I]:= X.V^[I];
end; {MultiUI.SetTo}

{--------------------------------------}
procedure MultiUI.Norm; {Normalize}
var
  I : Integer;
begin
  if N < 1 then  Clear;
  I:= N - 1; {Delete leading zeros}
  while (V^[I] = 0) and (I > 0) do begin
    Dec(I);
  end;
  N:= I + 1;
end; {MultiUI.Norm}

{--------------------------------------}
function MultiUI.AbComp( var X: MultiUI): Integer; {AbComp= sign(|Self| - |X|)}
{Si = -1 if |Self| < |X|; Si = 0 if |Self| = |X|; Si = +1 if |Self| > |X|}
{Self.S and X.S not used or changed;  Self, X not changed}
var
  I : Integer;
begin
  if N < X.N then begin  AbComp:= -1;  Exit;  end;
  if N > X.N then begin  AbComp:= +1;  Exit;  end;
  I:= N - 1;
  while (V^[I] = X.V^[I]) and (I > 0) do begin
    Dec(I);
  end;
  if V^[I] < X.V^[I] then begin  AbComp:= -1;  Exit;  end;
  if V^[I] > X.V^[I] then begin  AbComp:= +1;  Exit;  end;
  AbComp:= 0;
end; {MultiUI.AbComp}

{--------------------------------------}
procedure MultiUI.RAdd( var X : MultiUI); {Self:= Self + X}
{Sets MuErr True if overflow}
var
  K : Mu1Digit;
  I : Integer;
begin
  if N < X.N then begin  {If Digits in Self < Digits in X}
    if M >= X.N then begin
      for I:= N to X.N-1 do  V^[I]:= 0;
      N:= X.N;
    end
    else begin
      MuWriteErr('X too big to add, continuing...');
    end;
  end;
  IpAdd(X, Self, K);
  if (K > 0) then  MuWriteErr('Addition overflow, continuing...');
end; {MultiUI.RAdd}

{--------------------------------------}
procedure MultiUI.RSub( var X : MultiUI); {Self:= Self - X}
{Sets MuErr True if N > Self}
var
  I : Integer;
begin
  I:= AbComp(X);
  if I = 0 then  Clear
  else if I < 0 then         {If Self < X, Error}
    MuWriteErr('Unsigned subtraction error, continuing...')
  else
    IpSub(X, Self);
end; {MultiUI.RSub}

{--------------------------------------}
procedure MultiUI.Add( var X, Y : MultiUI); {Self:= X + Y}
{Sets MuErr True if overflow}
begin
  if      Self.V = X.V then
     RAdd(Y)
  else if Self.V = Y.V then
    RAdd(X)
  else if X.N >= Y.N then begin
    SetTo(X);  RAdd(Y);
  end
  else begin
    SetTo(Y);  RAdd(X);
  end;
end; {MultiUI.Add}

{--------------------------------------}
procedure MultiUI.Sub( var X, Y : MultiUI); {Self:= X - Y}
{Sets MuErr True if overflow}
var
  YL : ^MultiUI;
begin
  if @X = @Y then Clear
  else if Self.V = X.V then
    RSub(Y)
  else if Self.V = Y.V then begin
    New( YL, Init( Y.N));  YL^.SetTo(Y);
    SetTo(X);
    RSub( YL^);
    Dispose( YL, Done);
  end
  else begin
    SetTo(X);
    RSub(Y);
  end;
end; {MultiUI.Sub}

{--------------------------------------}
function MultiUI.ZTest : Boolean; {True if Self is Zero}
begin
  ZTest:= (N = 1) and (V^[0] = 0);
end; {MuLTI.ZTest}

{--------------------------------------}
function MultiUI.EQ1( D1 : Mu1Digit) : Boolean; {True if |Self| = D1}
{Self, D1 are not changed, allows literal D1}
begin
  EQ1:= (N = 1) and (V^[0] = D1);
end; {MultiUI.EQ1}

{--------------------------------------}
function MultiUI.GT1( D1 : Mu1Digit) : Boolean; {True if |Self| > D1}
{Self, D1 are not changed, allows literal D1}
begin
  GT1:= (N > 1) or (V^[0] > D1);
end; {MultiUI.GT1}

{------End of MultiUI's methods--------}

{--------------------------------------}
{   MultiSI's method implementation    }
{--------------------------------------}

{--------------------------------------}
constructor MultiSI.Init( NMaxI : Integer);
begin
  MultiUI.Init( NMaxI);
  S:= False;
end; {MultiSI.Init}

{--------------------------------------}
destructor MultiSI.Done;
begin
  MultiUI.Done;
end; {MultiSI.Done}

{--------------------------------------}
procedure MultiSI.Clear; {Set value = 0}
begin
  MultiUI.Clear;
  S:= False;
end; {MultiSI.Clear}

{--------------------------------------}
procedure MultiSI.ChS; {Change sign, Self = - Self}
begin
  if not ZTest then  S:= not S;
end; {MultiSI.ChS}

{--------------------------------------}
procedure MultiSI.SetTo( var X : MultiSI); {Self:= X}
begin
  MultiUI.SetTo(X);
  S:= X.S;
end; {MultiSI.SetTo}

{--------------------------------------}
procedure MultiSI.Norm; {Normalize}
begin
  MultiUI.Norm;
  if ZTest then  S:= False; {Prevent -0}
end; {MultiSI.Norm}

{--------------------------------------}
procedure MultiSI.RAdd( var X : MultiSI); {Self:= Self + X}
{Sets MuErr True if overflow}
var
  Temp : ^MultiUI;
  K : Mu1Digit;
  I : Integer;
begin
  if S = X.S then        {Same signs}
    MultiUI.RAdd(X)
  else begin             {Opposite signs}
    I:= AbComp(X);
    if I < 0 then begin  {If |Self| < |X|, |Self| = |X| - |Self|}
      New( Temp, Init(N));
      Temp^.SetTo( Self);  {Temp:= Small Self}
      Self.SetTo(X);      {Self:= Large X}
      IpSub( Temp^, Self);
      Dispose( Temp, Done);
    end
    else if I > 0 then   {If |Self| > |X|, |Self| = |Self| - |X|}
      IpSub(X, Self)
    else
      Clear;             {If |Self| = |X|, Self = 0}
  end;
end; {MultiSI.RAdd}

{--------------------------------------}
procedure MultiSI.RSub( var X : MultiSI); {Self:= Self - X}
{Sets MuErr True if overflow}
begin
  if Self.V = X.V then
    Clear
  else begin
    X.S:= not X.S;
    RAdd(X);
    X.S:= not X.S;
  end;
end; {MultiSI.RSub}

{--------------------------------------}
procedure MultiSI.Add( var X, Y : MultiSI); {Self:= X + Y}
{Sets MuErr True if overflow}
var
  I : Integer;
begin
  if      Self.V = X.V then
    RAdd(Y)
  else if Self.V = Y.V then
    RAdd(X)
  else if X.S = Y.S then begin
    MultiUI.Add(X, Y);  S:= X.S;
  end
  else begin
    I:= X.AbComp(Y);              {Signs are different}
    if I = 0 then  Clear
    else if I > 0 then begin
      SetTo(X);  IpSub(Y, Self);  {|X| > |Y|, so |X| - |Y|}
    end
    else begin
      SetTo(Y);  IpSub(X, Self);
    end;
  end;
end; {MultiSI.Add}

{--------------------------------------}
procedure MultiSI.Sub( var X, Y : MultiSI); {Self:= X - Y}
{Sets MuErr True if overflow}
begin
  if @X = @Y then Clear
  else if Self.V = X.V then
    RSub(Y)
  else if Self.V = Y.V then begin
    RSub(X);  ChS;
  end
  else begin
    Y.S:= not Y.S;
    Add(X, Y);
    Y.S:= not Y.S;
  end;
end; {MultiSI.Sub}

{--------------------------------------}
procedure MultiSI.Value( St : String; var I : Integer);
{Convert String to Self, returns I = # of character used}
{Allows St to be a literal, MuErr set True if overflow}
var
  K  : Mu1Digit;
  Len : Integer;
  J, D : Integer;
  Fini : Boolean;
  Overflow : Boolean;
begin
  I:= 1;
  While St[I] = ' ' do  Inc(I);
  if (St[I] = '-') or (St[I] = '+') then  Inc(I);
  while (I <= Length( St)) and
	(((St[I] >= '0') and (St[I] <= '9')) or (St[I] = ',')) do
    Inc(I);
  Dec(I);
  Clear;  Fini:= False;  Overflow:= False;
  Len:= I;
  if Len = 0 then  Exit;
  J:= 1;
  While St[J] = ' ' do  Inc(J); { HJS 91/04/24, 91/09/24 }
  if St[J] = '-' then begin  Inc(J);  S:= True;  end
  else if St[J] = '+' then  Inc(J);
  while Not Fini and (J <= Len) do begin
    if (St[J] >= '0') and (St[J] <= '9') then begin
      if (N = M) and (V^[ N-1] >= Base / 10) then begin
        Overflow:= True;
        Fini:= True;
      end
      else begin
        D:= Ord( St[J]) - Ord('0');
        IpMul1( 10, Self, K);
        if K > 0 then  Overflow:= True;
        V^[0]:= V^[0] + D;
      end;
    end
    else
      if St[J] <> ',' then  Fini:= True;
    Inc(J);
  end;
  if Overflow then  MuWriteErr('Input number overflow, continuing...');
  Norm;
end; {MultiSI.Value}

{--------------------------------------}
procedure MultiSI.SetTo1( D1 : Mu1Digit); {Self = D1 Mod Base}
{Allows a literal D1, D1 an integer, |D1| < Base, D1 not changed}
var
  AbsD : Mu1Digit;
begin
  AbsD:= Abs( D1);
  N:= 1;  S:= (D1 < 0);
  if AbsD < Base then  V^[0]:= AbsD
  else
    V^[0]:= AbsD - Int( AbsD / Base) * Base;
end; {MultiSI.SetTo1}

{--------------------------------------}
procedure MultiSI.SetToEx( Ex : Extended); {Self = Ex Mod Base**4}
{Allows a literal Ex, Ex not changed, MuErr set True if overflow}
{ Added procedure SetToEx HJS 91/05/02 }
var
  St1, St2   : String[ 255];
  AE, E1, E2 : Extended;
  L, ML      : Integer;
begin
  AE:= Abs( Ex);  E1:= Int( AE / Base);  E2:= AE - E1 * Base + Base;
  Str( E1:0:0, St1);  Str( E2:0:0, St2);
  Delete( St2, 1, 1);
  St1:= St1 + St2;
  L:= Length( St1);  ML:= 4 * MuDMax;
  if L > ML then  St1:= Copy( St1, L - ML + 1, ML);
  Value( St1, L);
  if Ex < 0.0 then ChS; {Prevents -0}
end; {MultiSI.SetToEx}

{--------------------------------------}
procedure MultiSI.SetToD( Db : Double); {Self = Db Mod Base**4}
{Allows a literal Db, Db not changed, MuErr set True if overflow}
begin
  SetToEx( Db);
end; {MultiSI.SetToD}

{--------------------------------------}
procedure MultiSI.Get1( var D1 : Mu1Digit); {D1 = Self Mod Base}
{Self is not changed}
begin
  D1:= V^[0];
  if S and (D1 <> 0) then  D1:= -D1;
end; {MultiSI.Get1}

{--------------------------------------}
procedure MultiSI.GetEx( var Ex : Extended); {Ex = Self Mod Base**4}
{Self is not changed}  { Added procedure GetEx HJS 91/05/02 }
var
  I : Integer;
  T : Extended;
begin
  I:= 1;  Ex:= V^[0];  T:= Base;
  while (I < N) and (I < 4) do begin
    Ex:= Ex + V^[I] * T; {Add next super digit}
    T:= T * Base;
    Inc(I);
  end;
  if S and (Ex <> 0.0) then  Ex:= -Ex;
end; {MultiSI.GetEx}

{--------------------------------------}
procedure MultiSI.GetD( var Db : Double); {Db = Self Mod Base**4}
{Self is not changed}
var
  T : Extended;
begin
  GetEx( T);  Db:= T;
end; {MultiSI.GetD}

{--------------------------------------}
procedure MultiSI.RAdd1( D1 : Mu1Digit); {Self = Self + D1}
{Allows literal D1, D1 is not changed, sets MuErr True if overflow}
var
  DL : ^MultiSI;
begin
  New( DL, Init(1));
  DL^.SetTo1( D1);
  RAdd( DL^);
  Dispose( DL, Done);
end; {MultiSI.RAdd1}

{--------------------------------------}
procedure MultiSI.Add1( var X : MultiSI; D1 : Mu1Digit); {Self = X + D1}
{Allows literal D1, D1 is not changed, sets MuErr True if overflow}
begin
  SetTo(X);
  RAdd1( D1);
end; {MultiSI.Add1}

{--------------------------------------}
function MultiSI.Comp( var X : MultiSI) : Integer; {Comp = sign( Self - X)}
{Sign = -1 if Self < X; Sign = 0 if Self = X; Sign = +1 if Self > X}
{Self.S and X.S used but not changed;  Self, X not changed}
begin
  if Not S and Not X.S then  Comp:= AbComp(X)       {Both positive}
  else if S and X.S    then  Comp:= X.AbComp( Self) {Both negative}
  else if X.S then  Comp:= +1                       {Self pos, X neg}
  else Comp:= -1;                                   {Self neg, X pos}
end; {MultiSI.Comp}

{--------------------------------------}
procedure WriteMax( var Out : TEXT; St : String);
{Output String and line feed every MuMaxW, MuTot = characters on line so far}
var
  I : Integer;
begin
  if (@Out = @OutPut) or (MuMaxW <= 0) then  Write( Out, St)
  else
    for I:= 1 to Length( St) do begin
      Write( Out, St[I]);  Inc( MuTot);
      if (MuTot >= MuMaxW) then begin
        WriteLn( Out);  MuTot:= 0;
      end;
    end;
end; {WriteMax}

{--------------------------------------}
procedure WriteMaxLn( var Out : TEXT; St : String);
{Output String and line feed every MuMaxW, MuTot = characters on line so far
 and a new line}
begin
  WriteMax( Out, St);
  WriteLn( Out);  MuTot:= 0;
end; {WriteMaxLn}

{--------------------------------------}
procedure MultiSI.Writ( var Out : TEXT); {Output Self as a line of text}
{Self is not modified}
var
  I, J, ToGo, LenD : LongInt;
  St : String[8];
begin
  ToGo:= LongInt(N) * MuDMax;
  if S then WriteMax( Out, '-');
  Str( Base + V^[N - 1] :0:0, St);
  J:= 1;
  repeat
    Inc(J);
  until (St[J] <> '0') or (J = MuDMax+1);
  WriteMax( Out, St[J]);
  Dec( ToGo, J - 1);
  LenD:= ToGo + 1;
  while J <= MuDMax do begin
    Inc(J);
    if (ToGo Mod MuDpG) = 0 then  WriteMax( Out, ',');
    WriteMax( Out, St[J]);
    Dec( ToGo);
  end;
  for I:= 2 to N do begin
    Str( Base + V^[N - I] :0:0, St);
    for J:= 2 to MuDMax+1 do begin
      if (ToGo Mod MuDpG) = 0 then  WriteMax( Out, ',');
      WriteMax( Out, St[J]);
      Dec( ToGo);
    end;
  end;
  if LenD >= MuLenD then begin
    Str( LenD, St);
    if (@Out <> @OutPut) and (MuMaxW > 0) and
       (Length(St) > (MuMaxW - MuTot - 3)) then begin
      WriteLn( Out);  MuTot:= 0;
    end;
    if (@Out = @OutPut) or (MuMaxW <= 0) or (MuTot <> 0) then
      WriteMax( Out, ' ');
    WriteMax( Out, '(' + St + ')');
  end;
end; {MultiSI.Writ}

{--------------------------------------}
procedure MultiSI.WritLn( var Out : TEXT); {Output Self and a new line}
begin
  Writ( Out);
  WriteLn( Out);  MuTot:= 0;
end; {MultiSI.WritLn}

{--------------------------------------}
procedure MultiSI.ShortWr( var Out : TEXT; Short : LongInt);
{Output Self in short form ... if more than Short digits}
{Self is not modified}
var
  Did : Integer;
  I, J, ToGo, LenD : LongInt;
  St : String[8];
begin
  if (N < 3) or (N <= Short div MuDMax) then begin
    Writ( Out);  Exit;
  end;
  ToGo:= LongInt(N) * MuDMax;
  if S then Write( Out, '-');
  Str( Base + V^[N - 1] :0:0, St);
  J:= 1;
  repeat
    Inc(J);
  until (St[J] <> '0') or (J = MuDMax+1);
  Write( Out, St[J]);  Did:= 1;
  Dec( ToGo, J - 1);
  LenD:= ToGo + 1;
  while (J <= MuDMax) and (Did < 5) do begin
    Inc(J);
    Write( Out, St[J]);  Inc( Did);
  end;
  Str( Base + V^[N - 2] :0:0, St);
  for J:= 2 to 6-Did do begin
    Write( Out, St[J]);
  end;
  Write( Out, ',...,');
  Str( Base + V^[0] :0:0, St);
  for J:= MuDMax-3 to MuDMax+1 do begin
    Write( Out, St[J]);
  end;
  Write( Out, ' (', LenD, ')');
end; {MultiSI.ShortWr}

{--------------------------------------}
procedure MultiSI.ShortWrLn( var Out: TEXT; Short : LongInt);
{Output Self short and a new line}
begin
  ShortWr( Out, Short);
  WriteLn( Out);  MuTot:= 0;
end; {MultiSI.ShortWrLn}

{--------------------------------------}
procedure MultiSI.RMul1( D1 : Mu1Digit); {Self = Self * D1}
{Multi-precision one super digit multiply; D1 is not changed}
{MuErr set true if Self overflows}
var
  K : Mu1Digit;
begin
  S:= S xor (D1 < 0);
  IpMul1(D1, Self, K);
  if K > 0 then
    MuWriteErr('One digit multiply overflow, continuing...');
end; {MultiSI.RMul1}

{--------------------------------------}
procedure MultiSI.Mul1( var X : MultiSI; D1 : Mu1Digit); {Self = X * D1}
{Multi-precision one super digit multiply; X and D1 are not changed}
{MuErr set true if Self overflows}
begin
  SetTo(X);
  RMul1( D1);
end; {MultiSI.Mul1}

{--------------------------------------}
procedure MultiSI.Mul( var X, Y : MultiSI); {Self = X * Y}
{X or Y or both may have the same address in memory as Self}
{X, Y are not changed unless they share memory with Self}
{Sets MuErr True if overflow}
var
  XL, YL   : ^MultiSI;
  I, J, IJ : Integer;
  K : Mu1Digit;
  T : Mu2Digit;
  NewXL, NewYL : Boolean;
begin
  XL:= @X;  NewXL:= False;
  YL:= @Y;  NewYL:= False;
  if XL^.N < YL^.N then begin
    XL:= @Y;  YL:= @X;
  end;
  if XL^.V = Self.V then begin
    New( XL, Init( Self.N));  XL^.SetTo( Self);  NewXL:= True;
  end
  else if YL^.V = Self.V then begin
    New( YL, Init( Self.N));  YL^.SetTo( Self);  NewYL:= True;
  end;
  if YL^.V = Self.V then  YL:= XL;  { HJS 1994-01-26 }
  S:= XL^.S xor YL^.S;
  for I:= 0 to XL^.N-1 do  V^[I]:= 0; {Clear lower of Self}
  KeyHit:= False;
  for J:= 0 to YL^.N-1 do begin
    K:= 0;  I:= 0;
    while I < XL^.N do begin
      IJ:= I + J;
      if IJ < Self.M then begin              {Mult, add and carry next Digit}
        T:= XL^.V^[I] * YL^.V^[J] + Self.V^[ IJ] + K;
        K:= Int(T / Base);
        Self.V^[ IJ]:= T - K * Base;
      end
      else
        I:= XL^.N; {Break out of inner loop}
      Inc(I);
    end;
    Inc( IJ);
    if IJ < Self.M then  Self.V^[ IJ]:= K;
    MuInterrupt;
    if MuAbort then begin
      Clear;
      if NewXL then  Dispose( XL, Done);
      if NewYL then  Dispose( YL, Done);
      Exit;
    end;
    if KeyHit then begin
      for I:= 0 to Echo do
        WriteLn( OutP[I]^, 'Integer Multiply: ',
          Trunc( (100.0 * (J+1)) / YL^.N), ' Percent done');
      KeyHit:= False;
    end;
  end;
  if K <> 0 then  Inc( IJ);
  if IJ > Self.M then begin
    IJ:= Self.M;
    MuWriteErr('Multiplication overflow, continuing...');
  end;
  N:= IJ;
  Norm;
  if NewXL then  Dispose( XL, Done);
  if NewYL then  Dispose( YL, Done);
end; {MultiSI.Mul}

{--------------------------------------}
procedure MultiSI.RMul( var X : MultiSI); {Self = Self * X}
{X may have the same address in memory as Self}
{X is not changed unless it shares memory with Self}
{Sets MuErr True if overflow}
begin
  Mul( Self, X);
end; {MultiSI.RMul}

{--------------------------------------}
procedure MultiSI.Sq( var X : MultiSI); {Self = X * X}
{X may have the same address in memory as Self}
{X is not changed unless it shares memory with Self}
{Sets MuErr True if overflow}
begin
  Mul(X, X);
end; {MultiSI.Sq}

{--------------------------------------}
procedure MultiSI.RSq; {Self = Self * Self}
{Sets MuErr True if overflow}
begin
  Mul( Self, Self);
end; {MultiSI.RSq}

{--------------------------------------}
procedure MultiSI.ModMB; {Self = Self Mod MuMB}
{Self is changed}
var  I : Integer;  QL : ^MultiSI;
begin
  if MuMB^.ZTest then  Exit;
  I:= AbComp( MuMB^);
  if I >= 0 then begin {if |Self| >= |MuMB|}
    New( QL, Init( Self.N));
    IpDiv( MuMB^, Self, QL^, Self);
    Dispose( QL, Done);
  end;
end; {MultiSI.ModMB}

{--------------------------------------}
procedure MultiSI.RDiv1( D1 : Mu1Digit); {Self = Self / D1}
{Multi-precision one super digit divide, D1 is not changed}
{Allows literal D1, MuErr set True if D1 = 0}
var
  RL : Mu1Digit;
begin
  IpDiv1( D1, Self, RL);
end; {MultiSI.RDiv1}

{--------------------------------------}
procedure MultiSI.Div1( var U : MultiSI; D1 : Mu1Digit); {Self = U / D1}
{Multi-precision one super digit divide, D1 and U not changed}
{U and Self may have same address in memory, then U is changed}
{Allows literal D1, MuErr set True if D1 = 0}
var
  RL : Mu1Digit;
begin
  SetTo(U);
  IpDiv1( D1, Self, RL);
end; {MultiSI.Div1}

{--------------------------------------}
procedure MultiSI.Mod1( D1: Mu1Digit; var R1: Mu1Digit); {R1 = Rem( Self / D1)}
{Multi-precision one super digit Modulo, D1 and Self not changed}
{D1 and R1 may have same address in memory, then D1 is changed}
{Allows literal D1, MuErr set True if D1 = 0}
var
  QL : ^MultiSI;
begin
  New( QL, Init( Self.N));
  QL^.SetTo( Self);
  IpDiv1( D1, QL^, R1);
  Dispose( QL, Done);
end; {MultiSI.Mod1}

{--------------------------------------}
procedure MultiSI.RDiv1QR( D1: Mu1Digit; var R1: Mu1Digit);
  {Self = Self / D1, R1 = Rem}
{Multi-precision one super digit Divide Q & R.  D1 and U not changed}
{D1 and R1 may have same address in memory, then D1 is changed}
{Allows literal D1, MuErr set True if D1 = 0}
begin
  IpDiv1( D1, Self, R1);
end; {MultiSI.RDiv1QR}

{--------------------------------------}
procedure MultiSI.Div1QR( var U: MultiSI; D1: Mu1Digit; var R1: Mu1Digit);
  {Self = U / D1, R1 = Rem}
{Multi-precision one super digit Divide Q & R.  D1 and U not changed}
{D1 and R1 may have same address in memory, then D1 is changed}
{Allows literal D1, MuErr set True if D1 = 0}
begin
  SetTo(U);
  IpDiv1( D1, Self, R1);
end; {MultiSI.Div1QR}

{--------------------------------------}
procedure MultiSI.Divi( var U, D : MultiSI); {Self = U / D}
{U or D or both may have the same address in memory as Self}
{U, D are not changed unless they share memory with Self}
{Sets MuErr True if D = 0}
var
  DL, RL : ^MultiSI;
  NewDL : Boolean;
begin
  DL:= @D;  NewDL:= False;
  if D.V = Self.V then begin
    New( DL, Init( D.N));  DL^.SetTo(D);  NewDL:= True;
  end;
  New( RL, Init( U.N));  RL^.SetTo(U);
  IpDiv( DL^, RL^, Self, RL^);
  if NewDL then  Dispose( DL, Done);
  Dispose( RL, Done);
end; {MultiSI.Divi}

{--------------------------------------}
procedure MultiSI.RDiv( var D : MultiSI); {Self = Self / D}
{D may have the same address in memory as Self}
{D is not changed unless it shares memory with Self}
{Sets MuErr True if D = 0}
begin
  Divi( Self, D);
end; {MultiSI.RDiv}

{--------------------------------------}
procedure MultiSI.Modu( var U, D : MultiSI); {Self = Rem(U / D)}
{U or D or both may have the same address in memory as Self}
{U, D are not changed unless they share memory with Self}
{Sets MuErr True if D = 0}
var
  DL, QL : ^MultiSI;
  NewDL : Boolean;
begin
  DL:= @D;  NewDL:= False;
  if D.V = Self.V then begin
    New( DL, Init( D.N));  DL^.SetTo(D);  NewDL:= True;
  end;
  New( QL, Init( U.N));
  SetTo(U);
  IpDiv( DL^, Self, QL^, Self);
  if NewDL then  Dispose( DL, Done);
  Dispose( QL, Done);
end; {MultiSI.Modu}

{--------------------------------------}
procedure MultiSI.RMod( var D : MultiSI); {Self = Rem( Self / D)}
{D may have the same address in memory as Self}
{D is not changed unless it shares memory with Self}
{Sets MuErr True if D = 0}
begin
  Modu( Self, D);
end; {MultiSI.RMod}

{--------------------------------------}
procedure MultiSI.DivQR( var U, D, R : MultiSI); {Self = U / D, R = Rem}
{U or D or both may have the same address in memory as Self or R}
{U, D are not changed unless they share memory with Self or R}
{Self and R may not have the same location in memory. Sets MuErr True if D = 0}
var
  DL : ^MultiSI;
  NewDL : Boolean;
begin
  if Self.V = R.V then begin
    MuWriteErr(
     'Cannot set same location to both quotient and remainder, continuing...');
    Exit;
  end;
  DL:= @D;  NewDL:= False;
  if (D.V = Self.V) or (@D =  @R) then begin
    New( DL, Init( D.N));  DL^.SetTo(D);  NewDL:= True;
  end;
  R.SetTo(U);
  IpDiv( DL^, R, Self, R);
  if NewDL then  Dispose( DL, Done);
end; {MultiSI.DivQR}

{--------------------------------------}
procedure MultiSI.RDivQR( var D, R : MultiSI); {Self = Self / D, R = Rem}
{Self may have the same address in memory as D or R}
{D is not changed unless it shares memory with Self or R}
{Sets MuErr True if D = 0}
begin
  DivQR( Self, D, R);
end; {MultiSI.RDivQR}

{--------------------------------------}
procedure MultiSI.PowMB( var B, P : MultiSI); {Self = (B ** P) Mod MuMB}
{B or P or both may have the same address in memory as Self}
{B, P are not changed unless they share memory with Self}
{Sets MuErr True if overflow or error}
var
  BL, PL : ^MultiSI;
  L      : LongInt;
  R      : Mu1Digit;
  Fini   : Boolean;
begin
  if B.ZTest then begin  {B = 0}
    if P.S then begin
      MuWriteErr('Cannot raise zero to a negative power');
    end
    else if P.ZTest then begin  {P = 0}
      MuWriteErr('Cannot raise zero to the zero power');
    end;
    Clear;  Exit;
  end;
  if (P.S) and ((B.N > 1) or (B.V^[0] > 1)) then begin { (>1) ** -P}
    Clear;  Exit;
  end;
  if (B.N = 1) and (B.V^[0] = 1) then begin {B = +/-1}
    L:= Round( P.V^[0]);
    S:= B.S and ((L Mod 2) = 1);
    N:= 1;  V^[0]:= 1;  Exit;
  end;

  Fini:= False;
  New( BL, Init( Self.M));  BL^.SetTo(B);
  New( PL, Init( P.N));     PL^.SetTo(P);
  SetTo1(1);
  repeat
    L:= Round( PL^.V^[0]);
    if (L Mod 2) = 1 then begin
      Self.Mul( BL^, Self);
      Self.ModMB;
    end;
    IpDiv1(2, PL^, R);
    if PL^.ZTest then Fini:= True
    else begin
      BL^.Mul( BL^, BL^);
      BL^.ModMB;
    end;
    if MuAbort then begin
      Clear;
      Fini:= True;
    end;
  until Fini;
  Dispose( PL, Done);
  Dispose( BL, Done);
end; {MultiSI.PowMB}

{--------------------------------------}
procedure MultiSI.RPowMB( var P : MultiSI); {Self = (Self ** P) Mod MuMB}
{P may have the same address in memory as Self}
{P is not changed unless it shares memory with Self}
{Sets MuErr True if overflow or error}
begin
  PowMB( Self, P);
end; {MultiSI.RPowMB}

{--------------------------------------}
procedure MultiSI.GCD( var A, B : MultiSI);{Self = Greatest common divisor A, B}
{A, B are not modified unless they share memory with Self}
{Will write to standard output if MuErrRep is True}
var
  AL, BL : ^MultiSI;
  I, NN  : Integer;
  Fini   : Boolean;
begin
  New( AL, Init( A.N));  AL^.SetTo(A);
  New( BL, Init( B.N));  BL^.SetTo(B);
  Fini:= False;
  AL^.S:= False;  BL^.S:= False;
  if AL^.ZTest then begin  Self.SetTo( BL^);  Fini:= True;  end;
  if BL^.ZTest then begin  Self.SetTo( AL^);  Fini:= True;  end;
  NN:= 0;
{ if MuErrRep and Not Fini then  for I:= 0 to Echo do  WriteLn( OutP[I]^); }
  while Not Fini do begin
    Inc( NN);
    if Odd( NN) then begin
      AL^.Modu( AL^, BL^);
{     if MuErrRep then begin } { HJS 92/02/29, two lines }
      if DiagOn then begin
        for I:= 0 to Echo do begin
	  Write( OutP[I]^, NN, ' ');  AL^.WritLn( OutP[I]^);
        end;
      end;
      if AL^.ZTest then begin  Self.SetTo( BL^);  Fini:= True;  end;
    end
    else begin
      BL^.Modu( BL^, AL^);
{     if MuErrRep then begin } { HJS 92/02/29, two lines }
      if DiagOn then begin
        for I:= 0 to Echo do begin
	  Write( OutP[I]^, NN, ' ');  BL^.WritLn( OutP[I]^);
        end;
      end;
      if BL^.ZTest then begin  Self.SetTo( AL^);  Fini:= True;  end;
    end;
  end;
  Dispose( AL, Done);
  Dispose( BL, Done);
end; {MultiSI.GCD}

{--------------------------------------}
procedure MultiSI.RGCD( var A: MultiSI); {Self=Greatest common divisor A, Self}
{A is not modified unless it shares memory with Self}
{Will write to standard output if MuErrRep is True}
begin
  GCD(A, Self);
end; {MultiSI.RGCD}

{--------------------------------------}
procedure MultiSI.FacMB( var X: MultiSI); {Self = (X !) Mod MuMB = 1*2*3*...*X}
{X may have the same address in memory as Self}
{X is not changed unless it shares memory with Self}
{Sets MuErr True if overflow or error}
var
  XL : ^MultiSI;
  NN : Double;
begin
  if X.S then begin
    MuWriteErr('Cannot take factorial of number < zero');
    Clear;  Exit;
  end;
  if X.N > 2 then begin
    MuWriteErr('Number too large for factorial function');
    Clear;  Exit;
  end;

  New( XL, Init( X.N));  XL^.SetTo(X);
  Self.SetTo1(1);
  XL^.GetD( NN);
  while NN > 1 do begin
    Self.Mul( XL^, Self);
    Self.ModMB;
    if MuAbort then begin
      Dispose( XL, Done);
      Clear;  Exit;
    end;
    NN:= NN - 1;
    XL^.SetToD( NN);
  end;
  Dispose( XL, Done);
end; {MultiSI.FacMB}

{--------------------------------------}
procedure MultiSI.RFacMB; {Self = (Self !) Mod MuMB = 1*2*3*...*X}
{Sets MuErr True if overflow or error}
begin
  FacMB( Self);
end; {MultiSI.RFacMB}

{--------------------------------------}
procedure MultiSI.SqRtRem( var X, R : MultiSI); {Self = SqRt(X), R = Rem}
{X is not changed, X is moved to R and then R is "divided"}
{in place to give the SqRt in Self and a remainder in R, it is ok for X and R}
{or X and Self to have the same location in memory, but then X will be changed}
{Self and R may not have the same location in memory. Sets MuErr True if error}
var
  D0, QH   : Mu1Digit;
  T, T2, D4, K  : Mu2Digit;
  I, J, RI0, DI0, RI, DI, IJ  : Integer;
begin
  if Self.V = R.V then begin
    MuWriteErr(
  'Cannot set same location to both square root and remainder, continuing...');
    Exit;
  end;
  R.SetTo(X);
  MultiSI.CLear;
  if R.S then begin
    MuWriteErr('Cannot take square root of negative number, continuing...');
    R.Clear;  Exit;
  end;
  if R.N < 4 then begin
    for I:= R.N to 3 do  R.V^[I]:= 0;
    R.N:= 4;
  end;
  if Odd( R.N) then begin
    R.V^[ R.N]:= 0;
    R.N:= R.N + 1;
  end;
  Self.N:= R.N div 2 + 1;
  T:= 0;
  for I:= 1 to 4 do  T:= T * Base + R.V^[ R.N - I];
  T2:= Int( SqRt(T));
  repeat
    D0:= Int( T2 / Base);
    QH:= 1 + T2 - D0 * Base; {+1 for safety}
    T:= R.V^[ R.N - 1] * Base + R.V^[ R.N - 2] - D0 * D0;
    if T < 0 then T2:= T2 - 1;
  until T >= 0;
  K:= Int(T / Base);
  R.V^[ R.N - 2]:= T - K * Base;
  R.V^[ R.N - 1]:= K;
  if Self.N >= 4 then Self.V^[ Self.N - 4]:= 0;
  Self.V^[ Self.N - 3]:= QH;
  T:= 2 * D0;
  K:= Int(T / Base);
  Self.V^[ Self.N - 2]:= T - K * Base;
  Self.V^[ Self.N - 1]:= K;
  KeyHit:= False;
  for J:= 2 to R.N div 2 do begin
    K:= Base;  RI0:= R.N - J - J;  DI0:= Self.N - J - 1;
    RI:= RI0;   DI:= DI0;   IJ:= RI0 + J;
    while RI <= IJ do begin     {Multiply and subtract}
      T:= R.V^[ RI] - QH * Self.V^[ DI] + K + BSMB; {+ BaseSq - Base}
      K:= Int(T / Base);
      T:= T - K * Base;
      R.V^[ RI]:= T;
      Inc( RI);  Inc( DI);
    end;
    T:= R.V^[ RI] + K + BSMB; {+ BaseSq - Base}
    K:= Int(T / Base);
    R.V^[ RI]:= T - K * Base;
    while K = MaxDigit do begin  {Test remainder}
{WriteLn('MuSqRt: Doing an add back'); .Diag for add back test}
      QH:= QH - 1;  {Add back}
      K:= QH;  RI:= RI0;  DI:= DI0;
      while RI <= IJ do begin
        T:= R.V^[ RI] + Self.V^[ DI] + K;
        K:= Int(T / Base);
        R.V^[ RI]:= T - K * Base;
        Inc( RI);  Inc( DI);
      end;
      T:= R.V^[ RI] + K;
      K:= Int(T / Base);
      R.V^[ RI]:= T - K * Base;
      K:= K + MaxDigit;
      Self.V^[ DI0]:= QH;
    end;
    I:= DI0;  IJ:= I + J;  K:= QH;
    while (I <= IJ) do begin
      T:= Self.V^[I] + K;         {Carry to next super digit}
      K:= Int(T / Base);
      Self.V^[I]:= T - K * Base;
      if K = 0 then  I:= IJ;  {Break out of loop}
      Inc(I);
    end;
    if J < R.N div 2 then begin
      T:= 0;
      for I:= 0 to 4 do begin
        T:= T * Base + R.V^[ R.N - J - I];
      end;
      D4:= 0;
      for I:= 1 to 4 do begin
        D4:= D4 * Base + Self.V^[ Self.N - I];
      end;
      QH:= 1 + Int(T / D4);
      Self.V^[ Self.N - J - 2]:= QH;
    end;
    MuInterrupt;
    if MuAbort then begin
      Self.Clear;  R.Clear;  Exit;
    end;
    if KeyHit then begin
      for I:= 0 to Echo do
        WriteLn( OutP[I]^, 'Integer Square Root: ',
          Trunc( (100.0 * J) / (R.N / 2)), ' Percent done');
      KeyHit:= False;
    end;
  end;
  R.N:= R.N div 2 + 1;
  R.Norm;
  Self.Norm;
  IpDiv1(2, Self, QH);
end; {MultiSI.SqRtRem}

{--------------------------------------}
procedure MultiSI.RSqRtRem( var R : MultiSI); {Self = SqRt( Self), R = Rem}
{Self is moved to R and then R is "divided" in place to give the SqRt in Self}
{and a remainder in R, Self and R may not have the same location in memory}
{Sets MuErr True if error}
begin
  SqRtRem( Self, R);
end; {MultiSI.RSqRtRem}

{--------------------------------------}
procedure MultiSI.SqRoot( var X : MultiSI); {Self = SqRt(X)}
{X is not changed, it is ok for X and Self to have the same location in memory}
{but then X will be changed.  Sets MuErr True if error}
var
  RL : ^MultiSI;
begin
  New( RL, Init( X.N));
  SqRtRem(X, RL^);
  Dispose( RL, Done);
end; {MultiSI.SqRoot}

{--------------------------------------}
procedure MultiSI.RSqRt; {Self = SqRt( Self)}
{Sets MuErr True if error}
begin
  SqRoot( Self);
end; {MultiSI.RSqRt}

{------End of MultiSI's methods--------}

{--------------------------------------}
{   End of method implementation       }
{--------------------------------------}

{--------------------------------------}
{         Other services               }
{--------------------------------------}

{--------------------------------------}
procedure ShiftUp( var St : String); {Shift string to upper case}
var
  I : Integer;
begin
  for I:= 1 to Length( St) do  St[I]:= UpCase( St[I]);
end; {ShiftUp}

{--------------------------------------}
procedure MuWriteErr( St : String); {Write error statement and set error flag}
{Allows St to be a literal string}
var
  I : Integer;
begin
  MuErr:= True;
  if MuErrRep then
    for I:= 0 to Echo do begin
      WriteLn( OutP[I]^, St);  WriteLn( OutP[I]^);
    end;
end; {MuWriteErr}

{--------------------------------------}
procedure ReadLInt( Mess  : String; Min, Max, Nom : LongInt;
                   var II : LongInt);   {Read in a long integer from keyboard}
var
  St   : String[ 255];
  Stat : Integer;
  LI   : LongInt;
  I    : Integer;
begin
  repeat
    repeat
      for I:= 0 to Echo do begin
        WriteLn( OutP[I]^, Mess);
        Write( OutP[I]^, '  [', Min, ', ', Max, '] (ENTER => ', Nom, '): ');
      end;
      ReadLn( St);
      for I:= 1 to Echo do  WriteLn( OutP[I]^, St);
    until IOResult = 0;
    Val( St, LI, Stat);
  until ((Stat = 0) and (LI >= Min) and (LI <= Max)) or (Length( St) = 0);
  if Length( St) = 0 then  LI:= Nom;
  II:= LI;
  for I:= 0 to Echo do begin
    WriteLn( OutP[I]^, 'Input = ', II);
    WriteLn( OutP[I]^);
  end;
end; {ReadLInt}

{--------------------------------------}
procedure ReadInt( Mess  : String; Min, Max, Nom : Integer;
                  var II : Integer);   {Read in an integer from keyboard}
var
  St   : String[ 255];
  Stat : Integer;
  LI   : LongInt;
  I    : Integer;
begin
  repeat
    repeat
      for I:= 0 to Echo do begin
        WriteLn( OutP[I]^, Mess);
        Write( OutP[I]^, '  [', Min, ', ', Max, '] (ENTER => ', Nom, '): ');
      end;
      ReadLn( St);
      for I:= 1 to Echo do  WriteLn( OutP[I]^, St);
    until IOResult = 0;
    Val( St, LI, Stat);
  until ((Stat = 0) and (LI >= Min) and (LI <= Max)) or (Length( St) = 0);
  if Length( St) = 0 then  LI:= Nom;
  II:= LI;
  for I:= 0 to Echo do begin
    WriteLn( OutP[I]^, 'Input = ', II);
    WriteLn( OutP[I]^);
  end;
end; {ReadInt}

{--------------------------------------}
procedure MuSetMMax( NRegs : Integer); {Set MMax}
var
  Max     : LongInt;
  Percent : Integer;
  I       : Integer;
begin
  Max:= MemAvail;
  for I:= 0 to Echo do
    WriteLn( OutP[I]^, 'MemAvail = ', Max);
  ReadInt('Input % of available memory to use:', 0, 100, 100, Percent);
  Max:= ((Max - 256) * Percent) div (100 * NRegs); {Max per register}
  MMax:= (Max - SizeOf( MultiSI)) div SizeOf( Mu1Digit) - 1;
  if MMax > MuNMax then  MMax:= MuNMax;
  if MMax < 5 then  MMax:= 5;
end; {MuSetMMax}

{--------------------------------------}
procedure MuInterrupt; {Test for ESC key pressed to abort operation}
{MuAbort set TRUE if ESC pressed twice, unchanged if no ESC or ESC, SPACE}
const
  Esc = Chr( 27); {ESCAPE}
  Ret = Chr( 13); {RETURN}
var
  I  : Integer;
  Ch : Char;
begin
  If not Keypressed then  Exit;
  while Keypressed do begin
    Ch:= ReadKey;
  end;
  KeyHit:= True;  KeyCh:= Ch;
  if Trace > 0 then begin
    for I:= 0 to Echo do
      WriteLn( OutP[I]^, '<', Tracn, ' ', Trace, '>');
    Trace:= 0;
  end;
  if (Ch <> Esc) or MuAbort then  Exit;
  for I:= 0 to Echo do begin
    WriteLn( OutP[I]^, '*** INTERRUPT:  To Continue Press RETURN Key;');
    WriteLn( OutP[I]^, 'To Abort Computation Press ESCAPE Key;');
    WriteLn( OutP[I]^, 'To Set SoftAbort Flag Press SPACE Bar.');
    WriteLn( OutP[I]^);
    KeyHit:= False;
  end;
  repeat
    Ch:= ReadKey;
  until (Ch = Esc) or (Ch = Ret) or (Ch = ' ');
  if Ch = Ret then  for I:= 0 to Echo do begin
    WriteLn( OutP[I]^, 'Continuing...');  WriteLn( OutP[I]^);
  end;
  if Ch = Esc then begin
    MuAbort:= True;
    for I:= 0 to Echo do begin
      WriteLn( OutP[I]^, 'Computation aborted by operator!');
      WriteLn( OutP[I]^);
    end;
  end;
  if Ch = ' ' then begin
    SoftAbort:= True;
    for I:= 0 to Echo do begin
      WriteLn( OutP[I]^, 'SoftAbort flag set by operator!');
      WriteLn( OutP[I]^);
    end;
  end;
end; {MuInterrupt}

{---------------------------------------}
function DosClock : Double; {Get the value of the Dos clock in total seconds}
{Only good to 0.01 seconds, must call earlier once each day}
var
  DC   : Double; {Local value of DosClock}
  Regs : Registers; {Registers for Dos calls defined in Turbo Pascal}
{ Regs : record
    case Integer of
    1 : (AX, BX, CX, DX, BP, SI, DI, DS, ES, Flags : Word);
    2 : (AL, AH, BL, BH, CL, CH, DL, DH : Byte);
    end;
}
begin
  Regs.AH:= $2C;
  MsDos( Regs);
  with Regs do
    DC:= DosDays * 86400.0 + CH * 3600.0 + CL * 60.0 + DH + DL / 100.0;
  if DC < DosClockP then begin
    Inc( DosDays);
    DC:= DC + 86400.0;
  end;
  DosClockP:= DC;
  DosClock := DC;
end; {DosClock}

{--------------------------------------}
procedure Diag( Mess : String); {Output a diagnostic message with delta times}
var
  EI   : Integer;
  Del0 : Double; {Total time from TV0}
  Del1 : Double; {Delta time from last Diag call}
begin
  if DiagOn then begin
    TV1:= TV2;  TV2:= DosClock;
    Del0:= TV2 - TV0;
    Del1:= TV2 - TV1;
    for EI:= 0 to Echo do
      WriteLn( OutP[ EI]^, 'T = ', Del0:0:2,
        '  DT = ', Del1:0:2, ' sec.  ',Mess);
  end;
end; {Diag}

{--------End of other services---------}

{--------------------------------------}
{        Initialization section        }
{--------------------------------------}

{Initialize Multi-precision package}
{Input: MuDMax, MuNMax}
{Output: Base, BaseSq, MaxDigit, BSMB, OutP}
{Initialize: MuDpG=5, MuLenD=21, MuErr=False, MuErrRep=True, MuAbort=False,}
{Echo=0, MMax=MuNMax,  TracN=0, Trace=0, DiagOn=False, TV0=TV1=TV2=DosClock}
var  I : Integer;
begin
  Base:= 10;
  for I:= 2 to MuDMax do  Base:= 10 * Base;
  MaxDigit := Base - 1;
  BaseSq   := Base * Base;
  BSMB     := BaseSq - Base;
  MuDpG    := 5;
  MuLenD   := 21;
  MuErr    := False;
  MuErrRep := True;
  MuAbort  := False;
  SoftAbort:= False;
  KeyHit   := False;
  KeyCh    := ' ';
  Echo     := 0;
  OutP[0]  := @Output;  OutP[1]:= @Lst;
  MMax     := MuNMax;
  TracN    := 0;
  Trace    := 0;
  DiagOn   := False;
  DosClockP:= 86400.0;
  DosDays  := 0;
  TV0      := DosClock;
  TV1      := TV0;
  TV2      := TV1;
  MuMaxW   := 65;
  MuTot    := 0;
end. {MultiIDO unit init}

Revisions made -
--------
Changed line in procedure MultiSI.Value from:
  While St[J] = ' ' do  J:= J + J;
to:
  While St[J] = ' ' do  Inc(J); { HJS 91/04/24, 91/09/24 }
--------
  Added procedures SetToEx and GetEx  { HJS 91/05/02 and 91/05/04 }
--------
  Added test of KeyHit in SqRtRem { HJS 91/08/17 }
--------
  Added MuMaxW and MuTot to global vars and MultiSI.Writ { HJS 92/02/07 }
--------
  Changed "if MuErrRep then begin"  { HJS 92/02/29 }
       to "if DiagOn then begin"  in MultiSI.GCD
  Added SoftAbort flag to MuInterrupt
--------
  In MultiSI.Mul( set XL & YL Local pointer to X or Y so XL^.N >= YL^.N
    { HJS 1993-01-11 }
--------
  In MultiSI.Mul( added  if YL^.V = Self.V then  YL:= XL;  { HJS 1994-01-26 }
    and removed  if @XL^ = @YL^ then  YL:= XL;
    to correct the case when YL^.V = Self.V and XL^.V = Self.V
--------

{End of file MultiID.Pas ***************************************************}
